# Copyright 2022 Camptocamp SA
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl)

import json
import os
import re
from collections import defaultdict

MANIFEST_NAMES = ("__manifest__.py", "__openerp__.py")


# Copy-pasted from OCA/maintainer-tools
def get_manifest_path(addon_dir):
    for manifest_name in MANIFEST_NAMES:
        manifest_path = os.path.join(addon_dir, manifest_name)
        if os.path.isfile(manifest_path):
            return manifest_path


class bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[39m"
    BOLD = "\033[1m"
    DIM = "\033[2m"
    ENDD = "\033[22m"
    UNDERLINE = "\033[4m"
    END = "\033[0m"


def clean_text(text):
    """Clean text by removing patterns like '13.0', '[13.0]' or '[IMP]'."""
    return re.sub(r"\[.*\]|\d+\.\d+", "", text).strip()


def defaultdict_from_dict(d):
    nd = lambda: defaultdict(nd)  # noqa
    ni = nd()
    ni.update(d)
    return ni


class Output:
    """Mixin to handle the output of oca-port."""

    def _print(self, *args, **kwargs):
        """Like built-in 'print' method but check if oca-port is used in CLI."""
        app = self
        # FIXME: determine class
        if hasattr(self, "app"):
            app = self.app
        if app.cli and not app.output:
            print(*args, **kwargs)

    def _render_output(self, output, data):
        """Render the data with the expected format."""
        return getattr(self, f"_render_output_{output}")(data)

    def _render_output_json(self, data):
        """Render the data as JSON."""
        return json.dumps(data)


class SmartDict(dict):
    """Dotted notation dict."""

    def __getattr__(self, attrib):
        val = self.get(attrib)
        return self.__class__(val) if isinstance(val, dict) else val


REF_REGEX = r"((?P<org>[\w-]+)/)?((?P<repo>[\w-]+)#)?(?P<branch>.*)"


def parse_gh_ref(ref):
    """Parse github reference in the form [org/][repo#]branch"""
    group = re.match(REF_REGEX, ref)
    return SmartDict(group.groupdict()) if group else None


def make_gh_info(kind, ref, remote=None):
    info = parse_gh_ref(ref)
    if not info:
        # FIXME
        raise ValueError(f"No valid {kind}")
    info["_kind"] = kind
    info["remote"] = remote or info.org
    if not info.org:
        info["org"] = "OCA"
    return info


def pr_ref_from_url(url):
    if not url:
        return ""
    # url like 'https://github.com/OCA/edi/pull/371'
    org, repo, __, nr = url.split("/")[3:]
    return f"{org}/{repo}#{nr}"



# Copyright 2022 Camptocamp SA
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl)
"""Tool helping to port an addon or missing commits of an addon from one branch
to another.

If the addon does not exist on the target branch, it will assist the user
in the migration, following the OCA migration guide.

If the addon already exists on the target branch, it will retrieve missing
commits to port. If a Pull Request exists for a missing commit, it will be
ported with all its commits if they were not yet (fully) ported.

To check if an addon could be migrated or to get eligible commits to port:

    $ export GITHUB_TOKEN=<token>
    $ oca-port 13.0 14.0 shopfloor --verbose

To effectively migrate the addon or port its commits, use the `--fork` option:

    $ oca-port 13.0 14.0 shopfloor --fork camptocamp


Migration of addon
------------------

The tool follows the usual OCA migration guide to port commits of an addon,
and will invite the user to fullfill the mentionned steps that can't be
performed automatically.

Port of commits/Pull Requests
-----------------------------

The tool will ask the user if he wants to open draft pull requests against
the upstream repository.

If there are several Pull Requests to port, it will ask the user if he wants to
base the next PR on the previous one, allowing the user to cumulate ported PRs
in one branch and creating a draft PR against the upstream repository with all
of them.
"""
import os

import click

from ..app import App
from ..exceptions import ForkValueError, RemoteBranchValueError
from ..utils.misc import bcolors as bc


@click.command()
@click.argument("from_branch", required=True)
@click.argument("to_branch", required=True)
@click.argument("addon", required=True)
@click.option(
    "--from-org",
    default="OCA",
    show_default=True,
    help="Upstream organization name.",
)
@click.option(
    "--from-remote",
    default="origin",
    show_default=True,
    required=True,
    help="Git remote from which source branches are fetched by default.",
)
@click.option("--repo-name", help="Repository name, eg. server-tools.")
@click.option(
    "--fork", help="Git remote where branches with ported commits are pushed."
)
@click.option("--user-org", show_default="--fork", help="User organization name.")
@click.option("--verbose", is_flag=True, help="List the commits of Pull Requests.")
@click.option(
    "--non-interactive", is_flag=True, help="Disable all interactive prompts."
)
@click.option(
    "--output",
    help=(
        "Returns the result in a given format. "
        "This implies the `--non-interactive` option automatically. "
        "Possibles values are: 'json'."
    ),
)
@click.option("--fetch", is_flag=True, help="Fetch remote branches from upstream.")
@click.option("--no-cache", is_flag=True, help="Disable user's cache.")
@click.option("--clear-cache", is_flag=True, help="Clear the user's cache.")
def main(
    from_branch: str,
    to_branch: str,
    addon: str,
    from_org: str,
    from_remote: str,
    repo_name: str,
    fork: str,
    user_org: str,
    verbose: bool,
    non_interactive: bool,
    output: str,
    fetch: bool,
    no_cache: bool,
    clear_cache: bool,
):
    """Migrate ADDON from FROM_BRANCH to TO_BRANCH or list Pull Requests to port
        if ADDON already exists on TO_BRANCH.

        Migration:

            Assist the user in the migration of the addon, following the OCA guidelines.

        Port of Pull Requests (missing commits):

            The PRs are found from FROM_BRANCH commits that do not exist in TO_BRANCH.
    The user will be asked if he wants to port them.

        To start the migration process, the `--fork` option must be provided in
    order to push the resulting branch on the user's remote.
    """
    try:
        app = App(
            addon=addon,
            from_branch=from_branch,
            to_branch=to_branch,
            from_org=from_org,
            from_remote=from_remote,
            repo_path=os.getcwd(),
            repo_name=repo_name,
            fork=fork,
            user_org=user_org,
            verbose=verbose,
            non_interactive=non_interactive,
            output=output,
            fetch=fetch,
            no_cache=no_cache,
            clear_cache=clear_cache,
            cli=True,
        )
    except ForkValueError as exc:
        error_msg = prepare_remote_error_msg(*exc.args)
        error_msg += (
            "\n\nYou can change the GitHub organization with the "
            f"{bc.DIM}--user-org{bc.END} option."
        )
        raise click.ClickException(error_msg) from exc
    except RemoteBranchValueError as exc:
        error_msg = prepare_remote_error_msg(*exc.args)
        raise click.ClickException(error_msg) from exc
    except ValueError as exc:
        raise click.ClickException(exc) from exc
    # Run the app
    try:
        app.run()
    except ValueError as exc:
        raise click.ClickException(exc) from exc


def prepare_remote_error_msg(repo_name, remote):
    return (
        f"No remote {bc.FAIL}{remote}{bc.END} in the current repository.\n"
        "To add it:\n"
        "\t# This mode requires an SSH key in the GitHub account\n"
        f"\t{bc.DIM}$ git remote add {remote} "
        f"git@github.com:{remote}/{repo_name}.git{bc.END}\n"
        "   Or:\n"
        "\t# This will require to enter user/password each time\n"
        f"\t{bc.DIM}$ git remote add {remote} "
        f"https://github.com/{remote}/{repo_name}.git{bc.END}"
    )


if __name__ == "__main__":
    main()



# Copyright 2022 Camptocamp SA
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl)

import re

import requests

from .git import PullRequest

GITHUB_API_URL = "https://api.github.com"


class GitHub:
    def __init__(self, token=None):
        self.token = token

    def request(self, url: str, method: str = "get", params=None, json=None):
        """Request GitHub API."""
        headers = {"Accept": "application/vnd.github.groot-preview+json"}
        if self.token:
            headers.update({"Authorization": f"token {self.token}"})
        full_url = "/".join([GITHUB_API_URL, url])
        kwargs = {"headers": headers}
        if json:
            kwargs.update(json=json)
        if params:
            kwargs.update(params=params)
        response = getattr(requests, method)(full_url, **kwargs)
        if not response.ok:
            raise RuntimeError(response.text)
        return response.json()

    def get_original_pr(
        self, from_org: str, repo_name: str, branch: str, commit_sha: str
    ):
        """Return original GitHub PR data of a commit."""
        gh_commit_pulls = self.request(
            f"repos/{from_org}/{repo_name}/commits/{commit_sha}/pulls"
        )
        gh_commit_pull = [
            data
            for data in gh_commit_pulls
            if (
                data["base"]["ref"] == branch
                and data["base"]["repo"]["full_name"] == f"{from_org}/{repo_name}"
            )
        ]
        return gh_commit_pull and gh_commit_pull[0] or {}

    def search_migration_pr(
        self, from_org: str, repo_name: str, branch: str, addon: str
    ):
        """Return an existing migration PR (if any) of `addon` for `branch`."""
        # NOTE: If the module we are looking for is named 'a_b' and the PR title is
        # written 'a b', we won't get any result, but that's better than returning
        # the wrong PR to the user.
        # NOTE 2: we first search for open PRs, then closed ones (could be closed
        # automatically by bots for inactivity)
        repo = f"{from_org}/{repo_name}"
        for pr_state in ("open", "unmerged"):
            prs = self.request(
                f"search/issues?q=is:pr+is:{pr_state}+repo:{repo}"
                f"+base:{branch}+in:title++mig+{addon}"
            )
            for pr in prs.get("items", {}):
                # Searching for 'a' on GitHub could return a result containing 'a_b'
                # so we check the result for the exact module name to return a relevant PR.
                if self._addon_in_text(addon, pr["title"]):
                    return PullRequest(
                        number=pr["number"],
                        url=pr["html_url"],
                        author=pr["user"]["login"],
                        title=pr["title"],
                        body=pr["body"],
                    )

    def _addon_in_text(self, addon: str, text: str):
        """Return `True` if `addon` is present in `text`."""
        return any(addon == term for term in re.split(r"\W+", text))



# Copyright 2023 Camptocamp SA
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl)

import io
import os
import pathlib
import sys
import tempfile
from collections import defaultdict
from contextlib import contextmanager

from oca_port.utils import misc


@contextmanager
def make_tmp_addon(name, manifest_name="__manifest__.py"):
    with tempfile.TemporaryDirectory() as path:
        os.system(f"mkdir -p {path}/addons/{name}")
        addons_path = f"{path}/addons"
        addon_path = f"{addons_path}/{name}"
        os.system(f"touch {addon_path}/{manifest_name}")
        yield pathlib.Path(addons_path)


@contextmanager
def capture_stdout():
    captured = io.StringIO()
    sys.stdout = captured
    yield captured
    sys.stdout = sys.__stdout__


def test_get_manifest_path():
    with make_tmp_addon("foo") as addons_path:
        expected = f"{addons_path}/foo/__manifest__.py"
        assert misc.get_manifest_path(addons_path / "foo") == expected


def test_clean_text():
    assert misc.clean_text("[13.0] foo 13.0") == "foo"


def test_default_dict_from_dict():
    res = misc.defaultdict_from_dict({"a": 1})
    # original values preserved
    assert res["a"] == 1
    # any key is a dict by default
    assert isinstance(res["b"], defaultdict)
    assert isinstance(res["b"]["c"], defaultdict)
    assert isinstance(res["b"]["c"]["d"], defaultdict)


def test_output_cli_mode():
    output = misc.Output()
    output.cli = True
    output.output = None
    with capture_stdout() as captured:
        output._print("foo")
        output._print("baz")
        output._print("bar")
        captured.seek(0)
        assert captured.read() == "foo\nbaz\nbar\n"

    assert output._render_output("json", {"a": 1}) == '{"a": 1}'



# Copyright 2022 Camptocamp SA
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl)

import os
import shutil
import tempfile
from collections import defaultdict

import click
import git

from .utils import git as g, misc
from .utils.misc import Output, bcolors as bc

AUTHOR_EMAILS_TO_SKIP = [
    "transbot@odoo-community.org",
    "noreply@weblate.org",
    "oca-git-bot@odoo-community.org",
    "oca+oca-travis@odoo-community.org",
    "oca-ci@odoo-community.org",
    "shopinvader-git-bot@shopinvader.com",
]

SUMMARY_TERMS_TO_SKIP = [
    "Translated using Weblate",
    "Added translation using Weblate",
]

PR_BRANCH_NAME = "oca-port-from-{from_branch}-to-{to_branch}-pr-{pr_number}"

FOLDERS_TO_SKIP = [
    "setup",
    ".github",
]

FILES_TO_KEEP = [
    "requirements.txt",
    "test-requirements.txt",
    "oca_dependencies.txt",
]

BOT_FILES_TO_SKIP = [
    "README.rst",
    "static/description/index.html",
]

# Fake PR for commits w/o any PR (used as fallback)
FAKE_PR = g.PullRequest(*[""] * 6)


def path_to_skip(commit_path):
    """Return True if the commit path should not be ported."""
    # Allows all folders (addons!) excepted those like 'setup/' generated
    # automatically by pre-commit.
    if commit_path.isdir:
        return commit_path in FOLDERS_TO_SKIP
    # Forbid all files excepted those that developers could update
    return commit_path not in FILES_TO_KEEP


class PortAddonPullRequest(Output):
    def __init__(self, app, create_branch=True, push_branch=True):
        """Port pull requests of an addon."""
        self.app = app
        self.create_branch = create_branch
        self.push_branch = push_branch
        self._results = {"process": "port_commits", "results": {}}

    def run(self):
        if not self.app.check_addon_exists_to_branch():
            if self.app.non_interactive:
                if self.app.output:
                    return False, self._render_output(self.app.output, {})
            return False, None
        self._print(
            f"{bc.BOLD}{self.app.addon}{bc.END} already exists "
            f"on {bc.BOLD}{self.app.to_branch.name}{bc.END}, "
            "checking PRs to port..."
        )
        branches_diff = BranchesDiff(self.app)
        branches_diff.print_diff(self.app.verbose)
        if self.app.non_interactive:
            if branches_diff.commits_diff:
                # If an output is defined we return the result in the expected format
                if self.app.output:
                    self._results["results"] = branches_diff.serialized_diff
                    return True, self._render_output(self.app.output, self._results)
                if self.app.cli:
                    # Exit with an error code if commits are eligible for (back)porting
                    # User-defined exit codes should be defined between 64 and 113.
                    # Allocate 110 for 'PortAddonPullRequest'.
                    raise SystemExit(110)
                return True, None
            if self.app.output:
                # Nothing to port -> return an empty output
                return False, self._render_output(self.app.output, {})
            return False, None
        if self.app.fork:
            self._print()
            self._port_pull_requests(branches_diff)
        return True, None

    def _port_pull_requests(self, branches_diff):
        """Open new Pull Requests (if it doesn't exist) on the GitHub repository."""
        base_ref = branches_diff.app.to_branch  # e.g. 'origin/14.0'
        previous_pr = previous_pr_branch = None
        processed_prs = []
        last_pr = (
            list(branches_diff.commits_diff.keys())[-1]
            if branches_diff.commits_diff
            else None
        )
        for pr, commits in branches_diff.commits_diff.items():
            current_commit = self.app.repo.commit(self.app.to_branch.ref())
            pr_branch, based_on_previous = self._port_pull_request_commits(
                pr,
                commits,
                base_ref,
                previous_pr,
                previous_pr_branch,
            )
            if pr_branch:
                # Check if commits have been ported.
                # If none has been ported, blacklist automatically the current PR.
                if self.app.repo.commit(pr_branch.ref()) == current_commit:
                    self._print("\tℹ️  Nothing has been ported, skipping")
                    self.app.storage.blacklist_pr(
                        pr.number,
                        confirm=True,
                        reason=f"(auto) Nothing to port from PR #{pr.number}",
                    )
                    if self.app.storage.dirty:
                        self.app.storage.commit()
                    msg = (
                        f"\t{bc.DIM}PR #{pr.number} has been"
                        if pr.number
                        else "Orphaned commits have been"
                    ) + f" automatically blacklisted{bc.ENDD}"
                    self._print(msg)
                    continue
                previous_pr = pr
                previous_pr_branch = pr_branch
                if based_on_previous:
                    processed_prs.append(pr)
                else:
                    processed_prs = [pr]
                if pr == last_pr:
                    self._print("\t🎉 Last PR processed! 🎉")
                is_pushed = self._push_branch_to_remote(pr_branch)
                if not is_pushed:
                    continue
                pr_data = self._prepare_pull_request_data(processed_prs, pr_branch)
                pr_url = self._search_pull_request(pr_data["base"], pr_data["title"])
                if pr_url:
                    self._print(f"\tExisting PR has been refreshed => {pr_url}")
                else:
                    self._create_pull_request(pr_branch, pr_data, processed_prs)

    def _port_pull_request_commits(
        self,
        pr,
        commits,
        base_ref,
        previous_pr=None,
        previous_pr_branch=None,
    ):
        """Port commits of a Pull Request in a new branch."""
        if pr.number:
            self._print(
                f"- {bc.BOLD}{bc.OKCYAN}Port PR #{pr.number}{bc.END} "
                f"({pr.url}) {bc.OKCYAN}{pr.title}{bc.ENDC}..."
            )
        else:
            self._print(f"- {bc.BOLD}{bc.OKCYAN}Port commits w/o PR{bc.END}...")
        based_on_previous = False
        # Ensure to not start to work from a working branch
        if self.app.to_branch.name in self.app.repo.heads:
            self.app.repo.heads[self.app.to_branch.name].checkout()
        else:
            self.app.repo.git.checkout(
                "--no-track",
                "-b",
                self.app.to_branch.name,
                self.app.to_branch.ref(),
            )
        # Ask the user if he wants to port the PR (or orphaned commits)
        if not click.confirm("\tPort it?" if pr.number else "\tPort them?"):
            self.app.storage.blacklist_pr(pr.ref, confirm=True)
            if not self.app.storage.dirty:
                return None, based_on_previous
        # Create a local branch based on upstream
        if self.create_branch:
            branch_name = PR_BRANCH_NAME.format(
                from_branch=self.app.from_branch.name,
                to_branch=self.app.to_branch.name,
                pr_number=pr.number,
            )
            if branch_name in self.app.repo.heads:
                # If the local branch already exists, ask the user if he wants
                # to recreate it + check if this existing branch is based on
                # the previous PR branch
                if previous_pr_branch:
                    based_on_previous = self.app.repo.is_ancestor(
                        previous_pr_branch.name, branch_name
                    )
                confirm = (
                    f"\tBranch {bc.BOLD}{branch_name}{bc.END} already exists, "
                    "recreate it?\n\t(⚠️  you will lose the existing branch)"
                )
                if not click.confirm(confirm):
                    return g.Branch(self.app.repo, branch_name), based_on_previous
                self.app.repo.delete_head(branch_name, "-f")
            if previous_pr and click.confirm(
                f"\tUse the previous {bc.BOLD}PR #{previous_pr.number}{bc.END} "
                "branch as base?"
            ):
                base_ref = previous_pr_branch
                based_on_previous = True
                # Set the new branch name the same than the previous one
                # but with the PR number as suffix.
                branch_name = f"{previous_pr_branch.name}-{pr.number}"
            self._print(
                f"\tCreate branch {bc.BOLD}{branch_name}{bc.END} from {base_ref.ref()}..."
            )
            self.app.repo.git.checkout("--no-track", "-b", branch_name, base_ref.ref())
        else:
            branch_name = self.app.to_branch.name
        # If the PR has been blacklisted we need to commit this information
        if self.app.storage.dirty:
            self.app.storage.commit()
            return g.Branch(self.app.repo, branch_name), based_on_previous

        # Cherry-pick commits of the source PR
        for commit in commits:
            self._print(
                f"\t\tApply {bc.OKCYAN}{commit.hexsha[:8]}{bc.ENDC} "
                f"{commit.summary}..."
            )
            # Port only relevant diffs/paths from the commit
            paths_to_port = set(commit.paths_to_port)
            for diff in commit.diffs:
                skip, message = self._skip_diff(commit, diff)
                if skip:
                    if message:
                        self._print(f"\t\t\t{message}")
                    if diff.a_path in paths_to_port:
                        paths_to_port.remove(diff.a_path)
                    if diff.b_path in paths_to_port:
                        paths_to_port.remove(diff.b_path)
                    continue
            if not paths_to_port:
                self._print("\t\t\tℹ️  Nothing to port from this commit, skipping")
                continue
            try:
                patches_dir = tempfile.mkdtemp()
                self.app.repo.git.format_patch(
                    "--keep-subject",
                    "-o",
                    patches_dir,
                    "-1",
                    commit.hexsha,
                    "--",
                    *paths_to_port,
                )
                patches = [
                    os.path.join(patches_dir, f)
                    for f in sorted(os.listdir(patches_dir))
                ]
                self.app.repo.git.am("-3", "--keep", *patches)
                shutil.rmtree(patches_dir)
            except git.exc.GitCommandError as exc:
                self._print(f"{bc.FAIL}ERROR:{bc.ENDC}\n{exc}\n")
                # High chance a conflict occurs, ask the user to resolve it
                if not click.confirm(
                    "⚠️  A conflict occurs, please resolve it and "
                    "confirm to continue the process (y) or skip this commit (N)."
                ):
                    self.app.repo.git.am("--abort")
                    continue
        return g.Branch(self.app.repo, branch_name), based_on_previous

    @staticmethod
    def _skip_diff(commit, diff):
        """Check if a commit diff should be skipped or not.

        A skipped diff won't have its file path ported through 'git format-path'.

        Return a tuple `(bool, message)` if the diff is skipped.
        """
        if diff.deleted_file:
            if diff.a_path not in commit.paths_to_port:
                return True, ""
        if diff.b_path not in commit.paths_to_port:
            return True, ""
        if diff.renamed:
            return False, ""
        diff_path = diff.b_path.split("/", maxsplit=1)[0]
        # Skip diff updating auto-generated files (pre-commit, bot...)
        if any(file_path in diff_path for file_path in BOT_FILES_TO_SKIP):
            return (
                True,
                f"SKIP: '{diff.change_type} {diff.b_path}' diff relates "
                "to an auto-generated file, skip to avoid conflict",
            )
        # Do not accept diff on unported addons
        if (
            not misc.get_manifest_path(diff_path)
            and diff_path not in commit.addons_created
        ):
            return (
                True,
                (
                    f"{bc.WARNING}SKIP diff "
                    f"{bc.BOLD}{diff.change_type} {diff.b_path}{bc.END}: "
                    "relates to an unported addon"
                ),
            )
        if diff.change_type in ("M", "D"):
            # Do not accept update and deletion on non-existing files
            if not os.path.exists(diff.b_path):
                return (
                    True,
                    (
                        f"SKIP: '{diff.change_type} {diff.b_path}' diff relates "
                        "to a non-existing file"
                    ),
                )
        return False, ""

    def _push_branch_to_remote(self, branch):
        """Force push the local branch to remote fork."""
        if not self.push_branch:
            return False
        confirm = (
            f"\tPush branch '{bc.BOLD}{branch.name}{bc.END}' "
            f"to remote '{bc.BOLD}{self.app.fork}{bc.END}'?"
        )
        if click.confirm(confirm):
            branch.repo.git.push(self.app.fork, branch.name, "--force-with-lease")
            branch.remote = self.app.fork
            return True
        return False

    def _prepare_pull_request_data(self, processed_prs, pr_branch):
        if len(processed_prs) > 1:
            title = (
                f"[{self.app.to_branch.name}][FW] {self.app.addon}: multiple ports "
                f"from {self.app.from_branch.name}"
            )
            lines = [f"- #{pr.number}" for pr in processed_prs]
            body = "\n".join(
                [
                    f"Port of the following PRs from {self.app.from_branch.name} "
                    f"to {self.app.to_branch.name}:"
                ]
                + lines
            )
        else:
            pr = processed_prs[0]
            title = f"[{self.app.to_branch.name}][FW] {pr.title}"
            body = (
                f"Port of #{pr.number} from {self.app.from_branch.name} "
                f"to {self.app.to_branch.name}."
            )
        return {
            "draft": True,
            "title": title,
            "head": f"{self.app.user_org}:{pr_branch.name}",
            "base": self.app.to_branch.name,
            "body": body,
        }

    def _search_pull_request(self, base_branch, title):
        params = {
            "q": (
                f"is:pr "
                f"repo:{self.app.from_org}/{self.app.repo_name} "
                f"base:{base_branch} "
                f"state:open {title} in:title"
            ),
        }
        response = self.app.github.request("search/issues", params=params)
        if response["items"]:
            return response["items"][0]["html_url"]

    def _create_pull_request(self, pr_branch, pr_data, processed_prs):
        if len(processed_prs) > 1:
            self._print(
                "\tPR(s) ported locally:",
                ", ".join(
                    [f"{bc.OKCYAN}#{pr.number}{bc.ENDC}" for pr in processed_prs]
                ),
            )
        if click.confirm(
            f"\tCreate a draft PR from '{bc.BOLD}{pr_branch.name}{bc.END}' "
            f"to '{bc.BOLD}{self.app.to_branch.name}{bc.END}' "
            f"against {bc.BOLD}{self.app.from_org}/{self.app.repo_name}{bc.END}?"
        ):
            response = self.app.github.request(
                f"repos/{self.app.from_org}/{self.app.repo_name}/pulls",
                method="post",
                json=pr_data,
            )
            pr_url = response["html_url"]
            self._print(
                f"\t\t{bc.BOLD}{bc.OKCYAN}PR created =>" f"{bc.ENDC} {pr_url}{bc.END}"
            )
            return pr_url


class BranchesDiff(Output):
    """Helper to compare easily commits (and related PRs) between two branches."""

    def __init__(self, app):
        self.app = app
        self.path = self.app.addon
        self.from_branch_path_commits, _ = self._get_branch_commits(
            self.app.from_branch.ref(), self.path
        )
        self.from_branch_all_commits, _ = self._get_branch_commits(
            self.app.from_branch.ref()
        )
        self.to_branch_path_commits, _ = self._get_branch_commits(
            self.app.to_branch.ref(), self.path
        )
        self.to_branch_all_commits, _ = self._get_branch_commits(
            self.app.to_branch.ref()
        )
        self.commits_diff = self.get_commits_diff()
        self.serialized_diff = self._serialize_diff(self.commits_diff)

    def _serialize_diff(self, commits_diff):
        data = {}
        for pr, commits in commits_diff.items():
            data[pr.number] = pr.to_dict()
            data[pr.number]["missing_commits"] = [commit.hexsha for commit in commits]
        return data

    def _get_branch_commits(self, branch, path="."):
        """Get commits from the local repository for the given `branch`.

        An optional `path` parameter can be set to limit commits to a given folder.
        This function also filters out undesirable commits (merge or translation
        commits...).

        Return two data structures:
            - a list of Commit objects `[Commit, ...]`
            - a dict of Commits objects grouped by SHA `{SHA: Commit, ...}`
        """
        commits = self.app.repo.iter_commits(branch, paths=path)
        commits_list = []
        commits_by_sha = {}
        for commit in commits:
            if self.app.cache.is_commit_ported(commit.hexsha):
                continue
            com = g.Commit(commit)
            if self._skip_commit(com):
                continue
            commits_list.append(com)
            commits_by_sha[commit.hexsha] = com
        # Put ancestors at the beginning of the list to loop with
        # the expected order
        commits_list.reverse()
        return commits_list, commits_by_sha

    @staticmethod
    def _skip_commit(commit):
        """Check if a commit should be skipped or not.

        Merge or translations commits are skipped for instance, or commits
        updating only files/folders we do not want to port (pre-commit
        configuration, setuptools files...).
        """
        return (
            # Skip merge commit
            len(commit.parents) > 1
            or commit.author_email in AUTHOR_EMAILS_TO_SKIP
            or any([term in commit.summary for term in SUMMARY_TERMS_TO_SKIP])
            or all(path_to_skip(path) for path in commit.paths)
        )

    def print_diff(self, verbose=False):
        lines_to_print = [""]
        fake_pr = None
        i = 0
        for i, pr in enumerate(self.commits_diff, 1):
            if pr.number:
                lines_to_print.append(
                    f"{i}) {bc.BOLD}{bc.OKBLUE}PR #{pr.number}{bc.END} "
                    f"({pr.url or 'w/o PR'}) {bc.OKBLUE}{pr.title}{bc.ENDC}:"
                )
                lines_to_print.append(f"\tBy {pr.author}, merged at {pr.merged_at}")
            else:
                lines_to_print.append(f"{i}) {bc.BOLD}{bc.OKBLUE}w/o PR{bc.END}:")
                fake_pr = pr
            if verbose:
                pr_paths = ", ".join([f"{bc.DIM}{path}{bc.ENDD}" for path in pr.paths])
                lines_to_print.append(f"\t=> Updates: {pr_paths}")
            if pr.number:
                pr_paths_not_ported = ", ".join(
                    [f"{bc.OKBLUE}{path}{bc.ENDC}" for path in pr.paths_not_ported]
                )
                lines_to_print.append(f"\t=> Not ported: {pr_paths_not_ported}")
            lines_to_print.append(
                f"\t=> {bc.BOLD}{bc.OKBLUE}{len(self.commits_diff[pr])} "
                f"commit(s){bc.END} not (fully) ported"
            )
            if verbose or not pr.number:
                for commit in self.commits_diff[pr]:
                    lines_to_print.append(
                        f"\t\t{bc.DIM}{commit.hexsha[:8]} " f"{commit.summary}{bc.ENDD}"
                    )
        if fake_pr:
            # We have commits without PR, adapt the message
            i -= 1
            nb_commits = len(self.commits_diff[fake_pr])
            message = (
                f"{bc.BOLD}{bc.OKBLUE}{i} pull request(s){bc.END} "
                f"and {bc.BOLD}{bc.OKBLUE}{nb_commits} commit(s) w/o "
                f"PR{bc.END} related to '{bc.OKBLUE}{self.path}"
                f"{bc.ENDC}' to port from {self.app.from_branch.ref()} "
                f"to {self.app.to_branch.ref()}"
            )
        else:
            message = (
                f"{bc.BOLD}{bc.OKBLUE}{i} pull request(s){bc.END} "
                f"related to '{bc.OKBLUE}{self.path}{bc.ENDC}' to port from "
                f"{self.app.from_branch.ref()} to {self.app.to_branch.ref()}"
            )
        lines_to_print.insert(0, message)
        self._print("\n".join(lines_to_print))

    def get_commits_diff(self):
        """Returns the commits which do not exist in `to_branch`, grouped by
        their related Pull Request.

        :return: a dict {PullRequest: {Commit: data, ...}, ...}
        """
        commits_by_pr = defaultdict(list)
        for commit in self.from_branch_path_commits:
            if commit in self.to_branch_all_commits:
                self.app.cache.mark_commit_as_ported(commit.hexsha)
                continue
            # Get related Pull Request if any
            pr = self._get_original_pr(commit)
            if pr:
                for pr_commit_sha in pr.commits:
                    try:
                        raw_commit = self.app.repo.commit(pr_commit_sha)
                    except ValueError:
                        # Ignore commits referenced by a PR but not present
                        # in the stable branches
                        continue
                    pr_commit = g.Commit(raw_commit)
                    if self._skip_commit(pr_commit):
                        continue
                    pr_commit_paths = {
                        path for path in pr_commit.paths if not path_to_skip(path)
                    }
                    pr.paths.update(pr_commit_paths)
                    # Check that this PR commit does not change the current
                    # addon we are interested in, in such case also check
                    # for each updated addons that the commit has already
                    # been ported.
                    # Indeed a commit could have been ported partially
                    # in the past (with git-format-patch), and we now want
                    # to port the remaining chunks.
                    if pr_commit not in self.to_branch_path_commits:
                        paths = set(pr_commit_paths)
                        # A commit could have been ported several times
                        # if it was impacting several addons and the
                        # migration has been done with git-format-patch
                        # on each addon separately
                        to_branch_all_commits = self.to_branch_all_commits[:]
                        skip_pr_commit = False
                        with g.no_strict_commit_equality():
                            while pr_commit in to_branch_all_commits:
                                index = to_branch_all_commits.index(pr_commit)
                                ported_commit = to_branch_all_commits.pop(index)
                                ported_commit_paths = {
                                    path
                                    for path in ported_commit.paths
                                    if not path_to_skip(path)
                                }
                                pr.ported_paths.update(ported_commit_paths)
                                pr_commit.ported_commits.append(ported_commit)
                                paths -= ported_commit_paths
                                if not paths:
                                    # The ported commits have already updated
                                    # the same addons than the original one,
                                    # we can skip it.
                                    skip_pr_commit = True
                        if skip_pr_commit:
                            continue
                    # We want to port commits that were still not ported
                    # for the addon we are interested in.
                    # If the commit has already been included, skip it.
                    if (
                        pr_commit in self.to_branch_path_commits
                        and pr_commit in self.to_branch_all_commits
                    ):
                        continue
                    existing_pr_commits = commits_by_pr.get(pr, [])
                    for existing_pr_commit in existing_pr_commits:
                        if (
                            existing_pr_commit == pr_commit
                            and existing_pr_commit.hexsha == pr_commit.hexsha
                        ):
                            # This PR commit has already been appended, skip
                            break
                    else:
                        commits_by_pr[pr].append(pr_commit)
            # No related PR: add the commit to the fake PR
            else:
                commits_by_pr[FAKE_PR].append(commit)
        # Sort PRs on the merge date (better to port them in the right order).
        # Do not return blacklisted PR.
        sorted_commits_by_pr = {}
        for pr in sorted(commits_by_pr, key=lambda pr: pr.merged_at or ""):
            blacklisted = self.app.storage.is_pr_blacklisted(pr.ref)
            if not blacklisted:
                # TODO: Backward compat for old tracking only by number
                blacklisted = self.app.storage.is_pr_blacklisted(pr.number)
            if blacklisted:
                msg = (
                    f"{bc.DIM}PR #{pr.number}" if pr.number else "Orphaned commits"
                ) + f" blacklisted ({blacklisted}){bc.ENDD}"
                self._print(msg)
                continue
            sorted_commits_by_pr[pr] = commits_by_pr[pr]
        return sorted_commits_by_pr

    def _get_original_pr(self, commit: g.Commit):
        """Return the original PR of a given commit."""
        # Try to get the data from the user's cache first
        data = self.app.cache.get_pr_from_commit(commit.hexsha)
        if data:
            return g.PullRequest(**data)
        # Request GitHub to get them
        if not any("github.com" in remote.url for remote in self.app.repo.remotes):
            return
        raw_data = self.app.github.get_original_pr(
            self.app.from_org,
            self.app.repo_name,
            self.app.from_branch.name,
            commit.hexsha,
        )
        if raw_data:
            # Get all commits of the PR as they could update others addons
            # than the one the user is interested in.
            # NOTE: commits fetched from PR are already in the right order
            pr_number = raw_data["number"]
            pr_commits_data = self.app.github.request(
                f"repos/{self.app.from_org}/{self.app.repo_name}"
                f"/pulls/{pr_number}/commits?per_page=100"
            )
            pr_commits = [pr["sha"] for pr in pr_commits_data]
            data = {
                "number": raw_data["number"],
                "url": raw_data["html_url"],
                "author": raw_data["user"].get("login", ""),
                "title": raw_data["title"],
                "body": raw_data["body"],
                "merged_at": raw_data["merged_at"],
                "commits": pr_commits,
            }
            self.app.cache.store_commit_pr(commit.hexsha, data)
            return g.PullRequest(**data)



# Copyright 2023 Camptocamp SA
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl)

import os

import click
import git

from ..exceptions import RemoteBranchValueError
from ..utils.git import Branch
from ..utils.storage import InputStorage


@click.group()
def cli():
    pass


@cli.command()
@click.argument("prs", required=True)
@click.argument("target_branch", required=True)
@click.argument("addon", required=True)
@click.option(
    "--reason",
    default="Nothing to port from PR #{pr_ref}",
    show_default=True,
)
@click.option(
    "--remote",
    default="origin",
    show_default=True,
)
def blacklist(
    prs: str,
    target_branch: str,
    addon: str,
    reason: str,
    remote: str,
):
    """Blacklist one or more PRs"""
    # eg: https://github.com/user/repo/pull/1234 or just the number
    # TODO: validate! Must be URL or ref like `OCA/edi#1`
    pr_refs = [x.strip() for x in prs.split(",") if x.strip()]

    # TODO: we assume you are in the right repo folder when you run this
    repo = git.Repo(os.getcwd())
    if repo.is_dirty(untracked_files=True):
        raise ValueError("changes not committed detected in this repository.")
    # Transform branch strings to Branch objects
    try:
        branch = Branch(repo, target_branch, default_remote=remote)
    except ValueError as exc:
        if exc.args[1] not in repo.remotes:
            raise RemoteBranchValueError(repo.name, exc.args[1]) from exc

    storage = InputStorage(branch, addon)
    for ref in pr_refs:
        storage.blacklist_pr(ref, reason=reason.format(pr_ref=ref))
    if storage.dirty:
        msg = f"oca-port: blacklist PR(s) {', '.join(pr_refs)} for {addon}"
        storage.commit(msg)


if __name__ == "__main__":
    cli()



# Copyright 2023 Camptocamp SA
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl)
{
    "name": "Test",
    "version": "1.0.0",
    "category": "Test Module",
    "author": "Camptocamp, Odoo Community Association (OCA)",
    "website": "https://github.com/OCA/oca-port",
    "license": "AGPL-3",
    "depends": ["base"],
    "data": [],
    "demo": [],
    "installable": True,
}



import json

from oca_port.app import App

from . import common


class TestApp(common.CommonCase):
    def _create_app(self, from_branch, to_branch, **kwargs):
        params = {
            "from_branch": from_branch,
            "to_branch": to_branch,
            "addon": self._settings["addon"],
            "from_org": self._settings["from_org"],
            "from_remote": self._settings["from_remote"],
            "repo_path": self.repo_path,
            "repo_name": "test",
            "user_org": self._settings["user_org"],
            "no_cache": self._settings["no_cache"],
        }
        params.update(kwargs)
        return App(**params)

    def test_app_nothing_to_port(self):
        app = self._create_app(
            self._settings["remote_branch1"], self._settings["remote_branch2"]
        )
        try:
            app.run()
        except SystemExit as exc:
            # exit code 0 means nothing needs to be migrated/ported
            self.assertEqual(exc.args[0], 0)

    def test_app_commit_to_port(self):
        self._commit_change_on_branch(
            self.repo_upstream_path, self._settings["branch1"]
        )
        app = self._create_app(
            self._settings["remote_branch1"],
            self._settings["remote_branch2"],
            fetch=True,
        )
        try:
            app.run()
        except SystemExit as exc:
            # exit code 110 means pull requests or commits could be ported
            self.assertEqual(exc.args[0], 110)
        # The other way around, no commit to backport (no exception)
        # (with CLI, the returned exit code is then 0)
        app = self._create_app(
            self._settings["remote_branch2"], self._settings["remote_branch1"]
        )
        res = app.run()
        self.assertFalse(res)

    def test_app_module_to_migrate(self):
        app = self._create_app(
            self._settings["remote_branch2"], self._settings["remote_branch3"]
        )
        try:
            app.run()
        except SystemExit as exc:
            # exit code 100 means the module could be migrated
            self.assertEqual(exc.args[0], 100)
        # The other way around, nothing to migrate as the module doesn't exist
        # (with CLI, the returned exit code is then 1)
        app = self._create_app(
            self._settings["remote_branch3"], self._settings["remote_branch2"]
        )
        error_msg = "my_module does not exist on origin/17.0"
        with self.assertRaisesRegex(ValueError, error_msg):
            app.run()

    def test_app_commit_to_port_non_interactive(self):
        self._commit_change_on_branch(
            self.repo_upstream_path, self._settings["branch1"]
        )
        app = self._create_app(
            self._settings["remote_branch1"],
            self._settings["remote_branch2"],
            non_interactive=True,
            fetch=True,
        )
        result = app.run()
        self.assertTrue(result)
        self.assertIsInstance(result, bool)
        # The other way around, no commit to backport
        app = self._create_app(
            self._settings["remote_branch2"],
            self._settings["remote_branch1"],
            non_interactive=True,
        )
        result = app.run()
        self.assertFalse(result)
        self.assertIsInstance(result, bool)

    def test_app_module_to_migrate_non_interactive(self):
        app = self._create_app(
            self._settings["remote_branch2"],
            self._settings["remote_branch3"],
            non_interactive=True,
        )
        result = app.run()
        self.assertTrue(result)
        self.assertIsInstance(result, bool)
        # The other way around, nothing to migrate as the module doesn't exist
        app = self._create_app(
            self._settings["remote_branch3"],
            self._settings["remote_branch2"],
            non_interactive=True,
        )
        error_msg = "my_module does not exist on origin/17.0"
        with self.assertRaisesRegex(ValueError, error_msg):
            app.run()

    def test_app_wrong_output(self):
        with self.assertRaisesRegex(ValueError, "Supported outputs are"):
            self._create_app(
                self._settings["remote_branch2"],
                self._settings["remote_branch3"],
                output="wrong_format",
            )

    def test_app_commit_to_port_output_json(self):
        commit_sha = self._commit_change_on_branch(
            self.repo_upstream_path, self._settings["branch1"]
        )
        app = self._create_app(
            self._settings["remote_branch1"],
            self._settings["remote_branch2"],
            output="json",
            fetch=True,
        )
        output = app.run()
        self.assertTrue(output)
        self.assertIsInstance(output, str)
        output = json.loads(output)
        self.assertEqual(output["process"], "port_commits")
        # A commit could be ported and is put in a "fake PR" without number
        self.assertEqual(len(output["results"]), 1)
        self.assertDictEqual(
            output["results"][""],
            {
                "url": "",
                "ref": "",
                "author": "",
                "title": "",
                "merged_at": "",
                "missing_commits": [commit_sha],
            },
        )
        # The other way around, no commit to backport
        app = self._create_app(
            self._settings["remote_branch2"],
            self._settings["remote_branch1"],
            output="json",
        )
        output = app.run()
        self.assertTrue(output)
        self.assertIsInstance(output, str)
        output = json.loads(output)
        self.assertEqual(output, {})

    def test_app_module_to_migrate_output_json(self):
        app = self._create_app(
            self._settings["remote_branch2"],
            self._settings["remote_branch3"],
            output="json",
        )
        output = app.run()
        self.assertTrue(output)
        self.assertIsInstance(output, str)
        output = json.loads(output)
        self.assertEqual(output["process"], "migrate")
        self.assertEqual(output["results"], {})
        # The other way around, nothing to migrate as the module doesn't exist
        app = self._create_app(
            self._settings["remote_branch3"],
            self._settings["remote_branch2"],
            output="json",
        )
        error_msg = "my_module does not exist on origin/17.0"
        with self.assertRaisesRegex(ValueError, error_msg):
            app.run()



import os
import pathlib
import shutil
import tempfile
import time
import unittest
from unittest.mock import patch

import git


class CommonCase(unittest.TestCase):
    _settings = {
        "branch1": "15.0",
        "remote_branch1": "origin/15.0",
        "branch2": "16.0",
        "remote_branch2": "origin/16.0",
        "branch3": "17.0",
        "remote_branch3": "origin/17.0",
        "addon": "my_module",
        "from_org": None,
        "from_remote": None,  # We're testing locally without any remote
        "user_org": "OCA-test",
        "no_cache": True,
    }

    def setUp(self):
        # Create a temporary Git repository
        self.repo_upstream_path = self._create_tmp_git_repository()
        self.module_path = os.path.join(
            self.repo_upstream_path, self._settings["addon"]
        )
        self.manifest_path = os.path.join(self.module_path, "__manifest__.py")
        self._fill_git_repository(self.repo_upstream_path)
        # By cloning the first repository this will set an 'origin' remote
        self.repo_path = self._clone_tmp_git_repository(self.repo_upstream_path)
        self._add_fork_remote(self.repo_path)
        # Patch GitHub class to prevent sending HTTP requests
        self._patch_github_class()

    def _patch_github_class(self):
        self.patcher = patch("oca_port.app.GitHub.request")
        github_request = self.patcher.start()
        github_request.return_value = {}
        self.addCleanup(self.patcher.stop)

    def _create_tmp_git_repository(self):
        """Create a temporary Git repository to run tests."""
        repo_path = tempfile.mkdtemp()
        git.Repo.init(repo_path)
        return repo_path

    def _clone_tmp_git_repository(self, upstream_path):
        repo_path = tempfile.mkdtemp()
        git.Repo.clone_from(upstream_path, repo_path)
        return repo_path

    def _fill_git_repository(self, repo_path):
        """Create branches with some content in the Git repository."""
        repo = git.Repo(repo_path)
        tpl_manifest_path = os.path.join(
            pathlib.Path(__file__).parent.resolve(),
            "data",
            "manifest.py",
        )
        with open(tpl_manifest_path) as tpl_manifest:
            tpl_manifest_lines = tpl_manifest.readlines()
        # Commit a file in '15.0'
        repo.git.checkout("--orphan", self._settings["branch1"])
        os.makedirs(self.module_path, exist_ok=True)
        with open(self.manifest_path, "w") as manifest:
            manifest.writelines(tpl_manifest_lines)
        repo.index.add(self.manifest_path)
        commit = repo.index.commit(f"[ADD] {self._settings['addon']}")
        # Port the commit from 'branch1' to 'branch2'
        repo.git.checkout("--orphan", self._settings["branch2"])
        repo.git.reset("--hard")
        # FIXME without a delay, both branches are targeting the same commit,
        # no idea why.
        time.sleep(1)
        repo.git.cherry_pick(commit.hexsha)
        # Create an empty 'branch3'
        repo.git.checkout("--orphan", self._settings["branch3"])
        repo.git.reset("--hard")
        repo.git.commit("-m", "Init", "--allow-empty")

    def _add_fork_remote(self, repo_path):
        repo = git.Repo(repo_path)
        # We do not really care about the remote URL here, re-use origin one
        repo.create_remote(self._settings["user_org"], repo.remotes.origin.url)

    def _commit_change_on_branch(self, repo_path, branch):
        """Commit a change that can be ported to another branch."""
        repo = git.Repo(repo_path)
        repo.git.checkout(branch)
        # Do some changes and commit
        with open(self.manifest_path, "r+") as manifest:
            content = manifest.read()
            content = content.replace('"base"', '"sale"')
            manifest.seek(0)
            manifest.write(content)
        repo.index.add(self.manifest_path)
        commit = repo.index.commit(f"[FIX] {self._settings['addon']}: fix dependency")
        return commit.hexsha

    def tearDown(self):
        # Clean up the Git repository
        shutil.rmtree(self.repo_upstream_path)
        shutil.rmtree(self.repo_path)



# Copyright 2022 Camptocamp SA
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl)
import os
import pathlib
from dataclasses import dataclass

import git

from . import utils
from .exceptions import ForkValueError, RemoteBranchValueError
from .migrate_addon import MigrateAddon
from .port_addon_pr import PortAddonPullRequest
from .utils.git import Branch
from .utils.github import GitHub
from .utils.misc import Output, bcolors as bc


@dataclass
class App(Output):
    """'oca-port' application centralizing settings and operations.

    Parameters:

        from_branch:
            the source branch (e.g. '15.0')
        to_branch:
            the source branch (e.g. '16.0')
        addon:
            the name of the module to process
        repo_path:
            local path to the Git repository
        fork:
            name of the Git remote used as fork
        repo_name:
            name of the repository on the upstream organization (e.g. 'server-tools')
        user_org:
            name of the user's GitHub organization where the fork is hosted
        from_org:
            name of the upstream GitHub organization (default = 'OCA')
        from_remote:
            name of the Git remote considered as the upstream (default = 'origin')
        verbose:
            returns more details to the user
        non_interactive:
            flag to not wait for user input and to return a error code to the shell.
            Returns 100 if an addon could be migrated, 110 if pull requests/commits
            could be ported, 0 if the history of the addon is the same on both branches.
        output:
            returns a parsable output. This implies the 'non-interactive' mode
            defined above but without returning any special exit code.
            Possible values: 'json'
        fetch:
            always fetch source and target branches from upstream
        no_cache:
            flag to disable the user's cache
        clear_cache:
            flag to remove the user's cache once the process is done
        github_token:
            Token to use when requesting GitHub API (highly recommended
            to not trigger the "API rate limit exceeded" error).
    """

    from_branch: str
    to_branch: str
    addon: str
    repo_path: str
    fork: str = None
    repo_name: str = None
    user_org: str = None
    from_org: str = "OCA"
    from_remote: str = "origin"
    verbose: bool = False
    non_interactive: bool = False
    output: str = None
    fetch: bool = False
    no_cache: bool = False
    clear_cache: bool = False
    github_token: str = None
    cli: bool = False  # Not documented, should not be used outside of the CLI

    _available_outputs = ("json",)

    def __post_init__(self):
        # Handle with repo_path and repo_name
        if self.repo_path:
            self.repo_path = pathlib.Path(self.repo_path)
        else:
            raise ValueError("'repo_path' has to be set.")
        if not self.repo_name:
            self.repo_name = self.repo_path.name
        # Handle Git repository
        self.repo = git.Repo(self.repo_path)
        if self.repo.is_dirty(untracked_files=True):
            raise ValueError("changes not committed detected in this repository.")
        # Handle user's organization and fork
        if not self.user_org:
            # Assume that the fork remote has the same name than the user organization
            self.user_org = self.fork
        if self.fork:
            if self.fork not in self.repo.remotes:
                raise ForkValueError(self.repo_name, self.fork)
        # Transform branch strings to Branch objects
        try:
            self.from_branch = Branch(
                self.repo, self.from_branch, default_remote=self.from_remote
            )
            self.to_branch = Branch(
                self.repo, self.to_branch, default_remote=self.from_remote
            )
        except ValueError as exc:
            if exc.args[1] not in self.repo.remotes:
                raise RemoteBranchValueError(self.repo_name, exc.args[1]) from exc
        # Force non-interactive mode:
        #   - if we are not in CLI mode
        if not self.cli:
            self.non_interactive = True
        #   - if an output has been defined
        if self.output:
            if self.output.lower() not in self._available_outputs:
                outputs = ", ".join(self._available_outputs)
                raise ValueError(f"Supported outputs are: {outputs}")
            self.non_interactive = True
        # Fetch branches if they can't be resolved locally
        # NOTE: required for the storage below to retrieve data
        remote_branches = self.repo.git.branch("-r").split()
        if (
            self.fetch
            or (
                self.from_branch.remote
                and self.from_branch.ref() not in remote_branches
            )
            or (self.to_branch.remote and self.to_branch.ref() not in remote_branches)
        ):
            self.fetch_branches()
        # GitHub API helper
        self.github = GitHub(self.github_token or os.environ.get("GITHUB_TOKEN"))
        # Initialize storage & cache
        self.storage = utils.storage.InputStorage(self.to_branch, self.addon)
        self.cache = utils.cache.UserCacheFactory(self).build()

    def fetch_branches(self):
        for branch in (self.from_branch, self.to_branch):
            if not branch.remote:
                continue
            remote_url = branch.repo.remotes[branch.remote].url
            if self.verbose:
                self._print(f"Fetch {bc.BOLD}{branch.ref()}{bc.END} from {remote_url}")
            branch.repo.remotes[branch.remote].fetch(branch.name)

    def _check_addon_exists(self, branch, raise_exc=False):
        repo = self.repo
        addon = self.addon
        branch_addons = [t.path for t in repo.commit(branch.ref()).tree.trees]
        if addon not in branch_addons:
            if not raise_exc:
                return False
            error = f"{addon} does not exist on {branch.ref()}"
            if self.cli:
                error = f"{bc.FAIL}{addon}{bc.ENDC} does not exist on {branch.ref()}"
            raise ValueError(error)
        return True

    def check_addon_exists_from_branch(self, raise_exc=False):
        """Check that `addon` exists on the source branch`."""
        return self._check_addon_exists(self.from_branch, raise_exc=raise_exc)

    def check_addon_exists_to_branch(self, raise_exc=False):
        """Check that `addon` exists on the target branch`."""
        return self._check_addon_exists(self.to_branch, raise_exc=raise_exc)

    def run(self):
        """Run 'oca-port' to migrate an addon or to port its pull requests."""
        self.check_addon_exists_from_branch(raise_exc=True)
        # Check if some PRs could be ported
        res, output = self.run_port()
        if not res:
            # If not, migrate the addon
            res, output = self.run_migrate()
        if self.cli and self.output:
            if not output:
                output = self._render_output(self.output, {})
            print(output)
        if self.clear_cache:
            self.cache.clear()
        if self.output:
            return output
        return res

    def run_port(self):
        """Port pull requests of an addon (if any)."""
        # Check if the addon (folder) exists on the target branch
        #   - if it already exists, check if some PRs could be ported
        return PortAddonPullRequest(self).run()

    def run_migrate(self):
        """Migrate an addon."""
        return MigrateAddon(self).run()



# Copyright 2023 Camptocamp SA
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl)

import unittest

from oca_port.utils import github


class TestGitHub(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.gh = github.GitHub(token="test")

    def test_token(self):
        assert self.gh.token == "test"

    def test_addon_in_text(self):
        # Matching OK
        res = self.gh._addon_in_text("a_b", "[16.0][MIG] a_b: migration to 16.0")
        assert res
        # Module name is not the expected one: do not match
        res = self.gh._addon_in_text("a_b", "[16.0][MIG] a_b_c: migration to 16.0")
        assert not res
